# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union
    from .snowflake import Snowflake


def calculate_shard_id(guild_id: Union[Snowflake, int], num_shards: int) -> int:
    """Calculates the shard receiving the events for a specified guild

    Parameters
    ----------
    guild_id : Optional[~pincer.utils.snowflake.Snowflake]
        The guild_id of the shard to look for
    num_shards : Optional[int]
        The number of shards.
    """
    return (guild_id >> 22) % num_shards



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild is deleted"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.guild import UnavailableGuild

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def guild_delete_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_guild_delete`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the guild delete event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.guild.UnavailableGuild`]
        ``on_guild_delete`` and an ``UnavailableGuild``
    """

    guild = UnavailableGuild.from_dict(payload.data)

    self.guilds.pop(guild.id, None)

    for channel in self.guild.channels:
        self.channels.pop(channel.id, None)

    return "on_guild_delete", guild


def export():
    return guild_delete_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when a user's voice state changes in a subscribed voice channel
(mute, volume, etc.)
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.user.voice_state import VoiceState

if TYPE_CHECKING:
    from typing import List, Tuple

    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def voice_state_update_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
) -> Tuple[str, List[VoiceState]]:
    """|coro|
    Middleware for the ``on_voice_state_update`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the voice state update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.user.voice_state.VoiceState`]
        ``on_voice_state_update`` and a ``VoiceState``
    """  # noqa: E501

    voice_state = VoiceState.from_dict(payload.data)
    guild = self.guilds.get(voice_state.guild_id)

    if guild:
        for index, state in enumerate(guild.voice_states):
            if state.user_id == voice_state.user_id:
                guild.voice_states[index] = voice_state
                break
        else:
            guild.voice_states.append(voice_state)

    return "on_voice_state_update", voice_state


def export():
    return voice_state_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from copy import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from ._component import _Component
from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List

    from ...objects.message.emoji import Emoji
    from ...utils.types import APINullable


@dataclass(repr=False)
class SelectOption(APIObject):
    """Represents a Discord Select Option

    Attributes
    ----------
    label: :class:`str`
        The user-facing name of the option, max 100 characters
    value: :class:`str`
        The def-defined value of the option, max 100 characters
    description: APINullable[:class:`str`]
        An additional description of the option, max 100 characters
    emoji: APINullable[:class:`~pincer.objects.message.emoji.Emoji`]
        ``id``, ``name``, and ``animated``
    default: APINullable[:class:`bool`]
        Will render this option as selected by default
    """
    label: str
    value: APINullable[str] = MISSING
    description: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    default: APINullable[bool] = MISSING

    def __post_init__(self):
        super().__post_init__()

        if self.value is MISSING:
            self.value = self.label


@dataclass(repr=False)
class SelectMenu(_Component):
    """Represents a Discord Select Menu

    Attributes
    ----------
    custom_id: :class:`str`
        A developer-defined identifier for the button,
        max 100 characters
    options: List[:class:`~pincer.commands.components.select_menu.SelectOption`]
        The choices in the select, max ``25``
    placeholder: APINullable[:class:`str`]
        Custom placeholder text if nothing is selected,
        max 100 characters
    min_values: APINullable[:class:`int`]
        The minimum number of items that must be chosen; min ``0``, max ``25``
        |default| ``1``
    max_values: APINullable[:class:`int`]
        The maximum number of items that can be chosen; max 25
        |default| ``1``
    disabled: APINullable[:class:`bool`]
        Disable the selects
        |default| False
    """
    custom_id: str
    options: Optional[List[SelectOption]] = None

    placeholder: APINullable[str] = MISSING
    min_values: APINullable[int] = 1
    max_values: APINullable[int] = 1
    disabled: APINullable[bool] = False

    type: int = 3

    def __post_init__(self):
        self.type = 3

    def with_options(self, *options: SelectOption) -> SelectMenu:
        """
        Sets the ``options`` parameter to \\*options and returns a new
        :class:`~pincer.commands.components.select_menu.SelectMenu`.

        \\*options : SelectOption
            List of options to set
        """
        copied_obj = copy(self)
        copied_obj.options = options
        return copied_obj

    def with_appended_options(self, *options: SelectOption) -> SelectMenu:
        """
        Append \\*options to the ``options`` parameter and returns a new
        :class:`~pincer.commands.components.select_menu.SelectMenu`.

        \\*options : SelectOption
            List of options to append
        """
        copied_obj = copy(self)
        copied_obj.options += options
        return copied_obj



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations
from asyncio import ensure_future

from importlib import reload, import_module
from inspect import isclass
from types import ModuleType
from typing import TYPE_CHECKING, List

from .commands.chat_command_handler import ChatCommandHandler
from .commands.interactable import Interactable
from .exceptions import CogAlreadyExists

if TYPE_CHECKING:
    from typing import Type
    from .client import Client


class CogManager:
    """
    A class that can load and unload cogs
    """

    def load_cog(self, cog: Type[Cog]):
        """Load a cog from a string path, setup method in COG may
        optionally have a first argument which will contain the client!

        :Example usage:

        run.py

        .. code-block:: python3

             from pincer import Client
             from cogs.say import SayCommand

             class MyClient(Client):
                 def __init__(self, *args, **kwargs):
                     self.load_cog(SayCommand)
                     super().__init__(*args, **kwargs)

        cogs/say.py

        .. code-block:: python3

             from pincer import command

             class SayCommand(Cog):
                 @command()
                 async def say(self, message: str) -> str:
                     return message

        Parameters
        ----------
        cog : Type[:class:`~pincer.cog.Cog`]
            The cog to load.
        """
        if cog in ChatCommandHandler.managers:
            raise CogAlreadyExists(
                f"Cog `{cog}` is trying to be loaded but already exists."
            )

        cog_manager = cog(self)

        ChatCommandHandler.managers.append(cog_manager)

    def load_cogs(self, *cogs: Type[Cog]):
        """
        Loads a list of cogs

        Parameters
        ----------
        \\*cogs : Type[:class:`~pincer.cog.Cog`]
            A list of cogs to load.
        """
        for cog in cogs:
            self.load_cog(cog)

    def load_module(self, module: ModuleType):
        """Loads the cogs from a module recursively.

        Parameters
        ----------
        module : :class:`~types.ModuleType`
            The module to load.
        """
        for item in vars(module).values():
            if isinstance(item, ModuleType):
                self.load_module(item)
            elif item is not Cog and isclass(item) and issubclass(item, Cog):
                self.load_cog(item)

    def reload_cogs(self):
        """Reloads all of the loaded cogs"""

        modules = []

        for cog in self.cogs:
            cog.unassign()

            mod = import_module(type(cog).__module__)
            if mod not in modules:
                modules.append(mod)

        for mod in modules:
            reload(mod)

        for cog in self.cogs:
            for mod in modules:
                cog = getattr(mod, type(cog).__name__, None)
                if cog:
                    self.load_cog(cog)

        ChatCommandHandler.has_been_initialized = False
        ensure_future(ChatCommandHandler(self).initialize())

    @property
    def cogs(self) -> List[Cog]:
        """Get a dictionary of all loaded cogs.

        The key/value pair is import path/cog class.

        Returns
        -------
        List[:class:`~pincer.cog.Cog`]
            The list of cogs
        """
        return [
            manager for manager in ChatCommandHandler.managers
            if isinstance(manager, Cog)
        ]


class Cog(Interactable):
    """A cog object
    This is an object that can register commands and message components that isn't a
    client. It also can be loaded and unloaded at runtime so commands can be changed
    without restarting the bot.
    """

    def __init__(self, client: Client) -> None:
        self.client = client

        super().__init__()

    @classmethod
    def name(cls) -> str:
        """
        Returns a unique name for this cog.

        Returns
        -------
        str
            A unique name for this cog.
        """
        return f"{cls.__module__}.{cls.__name__}"



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""Sent when a stage instance is created."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects import StageInstance
from ..utils.types import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def stage_instance_create_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_stage_instance_create`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the stage instance create event
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.stage.StageInstance`]
        ``on_stage_instance_create`` and a ``StageInstance``
    """

    stage = StageInstance.from_dict(payload.data)

    guild = self.guilds.get(stage.guild_id)
    if guild:
        guild.stage_instances.append(stage)

    return "on_stage_instance_create", stage


def export() -> Coro:
    return stage_instance_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""Event sent when a subscribed server's state changes"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.guild import GuildStatusEvent
from ..utils.types import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def guild_status_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_guild_status`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the guild status event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Return
    ------
    Tuple[:class:`str`, :class:`~pincer.objects.events.guild.GuildStatusEvent`]
        ``on_guild_status`` and a ``GuildStatusEvent``
    """
    return ("on_guild_status", GuildStatusEvent.from_dict(payload.data))


def export() -> Coro:
    return guild_status_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild emoji is updated."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.guild import GuildEmojisUpdateEvent
from ..utils import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def guild_emojis_update_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_guild_emojis_update`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the guild emojis update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.guild.GuildEmojisUpdateEvent`]
        ``on_guild_emoji_update`` and a ``GuildEmojisUpdateEvent``
    """  # noqa: E501

    event = GuildEmojisUpdateEvent.from_dict(payload.data)
    guild = self.guild.get(event.guild_id)

    if guild:
        guild.emojis = event.emojis

    return ("on_guild_emojis_update", event)


def export() -> Coro:
    return guild_emojis_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .audit_log import (
    AuditLogEvent, AuditLogChange, AuditEntryInfo, AuditLogEntry, AuditLog
)
from .ban import Ban
from .channel import (
    ChannelType, Channel, TextChannel, VoiceChannel, CategoryChannel,
    NewsChannel, ChannelMention, PrivateThread, PublicThread
)
from .features import GuildFeature
from .followed_channel import FollowedChannel
from .guild import (
    PremiumTier, GuildNSFWLevel, ExplicitContentFilterLevel, MFALevel,
    VerificationLevel, DefaultMessageNotificationLevel, SystemChannelFlags,
    Guild, UnavailableGuild
)
from .invite import (
    InviteTargetType, InviteStageInstance, Invite
)
from .member import GuildMember, PartialGuildMember, BaseMember
from .overwrite import Overwrite
from .permissions import Permissions
from .role import RoleTags, Role
from .scheduled_events import GuildScheduledEventEntityType, GuildScheduledEventUser, EventStatus, ScheduledEvent
from .stage import PrivacyLevel, StageInstance
from .template import GuildTemplate
from .thread import ThreadMetadata, ThreadMember
from .webhook import WebhookType, Webhook
from .welcome_screen import WelcomeScreenChannel, WelcomeScreen
from .widget import GuildWidget


__all__ = (
    "", "AuditEntryInfo", "AuditLog", "AuditLogChange",
    "AuditLogEntry", "AuditLogEvent", "Ban", "BaseMember", "CategoryChannel",
    "Channel", "ChannelMention", "ChannelType",
    "DefaultMessageNotificationLevel", "EventStatus",
    "ExplicitContentFilterLevel", "FollowedChannel", "Guild", "GuildFeature",
    "GuildMember", "GuildNSFWLevel", "GuildScheduledEventEntityType",
    "GuildTemplate", "GuildWidget", "Invite", "InviteStageInstance",
    "InviteTargetType", "MFALevel", "NewsChannel", "Overwrite",
    "PartialGuildMember", "Permissions", "PremiumTier", "PrivacyLevel", "Role",
    "RoleTags", "ScheduledEvent", "StageInstance", "SystemChannelFlags",
    "TextChannel", "ThreadMember", "ThreadMetadata", "UnavailableGuild",
    "VerificationLevel", "VoiceChannel", "Webhook", "WebhookType",
    "WelcomeScreen", "WelcomeScreenChannel"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from inspect import getfullargspec, Parameter, Signature
from typing import Any, Union, Callable, Mapping, List

from .types import Coro, TypeCache
from ..objects.message.context import MessageContext


def should_pass_cls(call: Union[Coro, Callable[[Any], Any]]) -> bool:
    """
    Checks whether a callable requires a self/cls as first parameter.

    Parameters
    ----------
    call: Union[Coro, Callable[[Any], Any]]
        The callable to check.

    Returns
    -------
    bool
        Whether the callable requires a self/cls as first parameter.
    """
    args = getfullargspec(call).args
    return len(args) >= 1 and args[0] in ["self", "cls"]


def should_pass_gateway(call: Union[Coro, Callable[[Any], Any]]) -> bool:
    """
    Checks whether a callable requires a dispatcher as last parameter.

    Parameters
    ----------
    call: Union[:class:`~pincer.utils.types.Coro`, Callable[[Any], Any]]
        The callable to check.

    Returns
    -------
    bool
        Whether the callable requires a dispatcher as first parameter.
    """
    args = getfullargspec(call).args
    return len(args) >= 2 and args[1] in ("gateway", "shard")


context_types = [Signature.empty, MessageContext]


def should_pass_ctx(sig: Mapping[str, Parameter], params: List[str]) -> bool:
    # TODO: Fix docs
    """

    Parameters
    ----------
    sig
    params

    Returns
    -------

    """
    if not params:
        return False

    annotation = sig[params[0]].annotation
    if isinstance(annotation, str):
        TypeCache()
        annotation = eval(annotation, TypeCache.cache, globals())

    return (
        annotation == MessageContext
        or params[0] == "ctx"
    )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional


@dataclass
class Group:
    """
    The group object represents a group that commands can be in. This is always a top
    level command.

    .. code-block:: python

        class Bot:

            group = Group("cool_commands")

            @command(parent=group)
            async def a_very_cool_command():
                pass

    This code creates a command called ``cool_commands`` with the subcommand
    ``a_very_cool_command``

    Parameters
    ----------
    name : str
        The name of the command group.
    description : Optional[:class:`str`]
        The description of the command. This has to be sent to Discord, but it does
        nothing, so it is optional.
    """
    name: str
    description: Optional[str] = None

    def __hash__(self) -> int:
        return hash(self.name)


@dataclass
class Subgroup:
    """
    A subgroup of commands. This allows you to create subcommands inside a
    subcommand-group.

    .. code-block:: python

        class Bot:

            group = Group("cool_commands")
            sub_group = Subgroup("group_of_cool_commands")

            @command(parent=sub_group)
            async def a_very_cool_command():
                pass

    This code creates a command called ``cool_commands`` with the subcommand-group
    ``group_of_cool_commands`` that has the subcommand ``a_very_cool_command``.

    Parameters
    ----------
    name : str
        The name of the command sub-group.
    parent : :class:`~pincer.commands.groups.Group`
        The parent group of this command.
    description : Optional[:class:`str`]
        The description of the command. This has to be sent to Discord, but it does
        nothing, so it is optional.
    """
    name: str
    parent: Group
    description: Optional[str] = None

    def __hash__(self) -> int:
        return hash(self.name)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import logging
from copy import copy
from contextlib import suppress
from inspect import isasyncgenfunction, _empty
from typing import TYPE_CHECKING, Optional

from ..commands import ChatCommandHandler, ComponentHandler
from ..commands.chat_command_handler import _hash_app_command_params
from ..exceptions import InteractionDoesNotExist
from ..objects import (
    Interaction,
    MessageContext,
    AppCommandType,
    InteractionType,
)
from ..utils import MISSING, should_pass_cls, Coro, should_pass_ctx
from ..utils import get_index
from ..utils.signature import get_signature_and_params

if TYPE_CHECKING:
    from typing import Any, Dict, List, Tuple
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch

_log = logging.getLogger(__name__)


def get_command_from_registry(interaction: Interaction):
    """
    Search for a command in ChatCommandHandler.register and return it if it exists.
    The naming of commands is converted from the Discord version to the Pincer version
    before checking the cache. See ChatCommandHandler docs for more information.

    Parameters
    ---------
    interaction : :class:`~pincer.objects.app.interactions.Interaction`
        The interaction to get the command from

    Raises
    ------
    :class:`~pincer.exceptions.InteractionDoesNotExist`
        The command is not registered
    """

    name: str = interaction.data.name
    group = None
    sub_group = None

    options = interaction.data.options

    if interaction.data.options:
        option = options[0]
        if option.type == 1:
            group = name
            name = option.name
        elif option.type == 2:
            group = interaction.data.name
            sub_group = option.name
            name = option.options[0].name

    with suppress(KeyError):
        return ChatCommandHandler.register[
            _hash_app_command_params(
                name, MISSING, interaction.data.type, group, sub_group
            )
        ]

    with suppress(KeyError):
        return ChatCommandHandler.register[
            _hash_app_command_params(
                name,
                interaction.guild_id,
                interaction.data.type,
                group,
                sub_group,
            )
        ]

    raise InteractionDoesNotExist(
        f"No command is registered for {interaction.data.name} with type"
        f" {interaction.data.type}"
    )


def get_call(self: Client, interaction: Interaction) -> Optional[Tuple[Coro, Any]]:
    if interaction.type == InteractionType.APPLICATION_COMMAND:
        command = get_command_from_registry(interaction)
        if command is None:
            return None
        # Only application commands can be throttled
        self.throttler.handle(command)
        return command.call, command.manager
    elif interaction.type == InteractionType.MESSAGE_COMPONENT:
        command = ComponentHandler.register.get(interaction.data.custom_id)
        return command.call, command.manager
    elif interaction.type == InteractionType.AUTOCOMPLETE:
        raise NotImplementedError(
            "Handling for autocomplete is not implemented"
        )


async def interaction_response_handler(
    self: Client,
    command: Coro,
    manager: Any,
    context: MessageContext,
    interaction: Interaction,
    args: List[Any],
    kwargs: Dict[str, Any],
):
    """|coro|

    Handle any coroutine as a command.

    Parameters
    ----------
    command : :class:`~pincer.utils.types.Coro`
        The coroutine which will be seen as a command.
    context : :class:`~pincer.objects.message.context.MessageContext`
        The context of the command.
    interaction : :class:`~pincer.objects.app.interactions.Interaction`
        The interaction which is linked to the command.
    \\*\\*kwargs :
        The arguments to be passed to the command.
    """
    # Prevent args from being mutated unexpectedly
    args = copy(args)

    if should_pass_ctx(*get_signature_and_params(command)):
        args.insert(0, context)

    if should_pass_cls(command):
        args.insert(0, manager or self)

    if isasyncgenfunction(command):
        message = command(*args, **kwargs)

        async for msg in message:
            if interaction.has_replied:
                await interaction.followup(msg)
            else:
                await interaction.reply(msg)
    else:
        message = await command(*args, **kwargs)
        if not interaction.has_replied:
            await interaction.reply(message)


async def interaction_handler(
    self: Client,
    interaction: Interaction,
    context: MessageContext,
    command: Coro,
    manager: Any
):
    """|coro|

    Processes an interaction.

    Parameters
    ----------
    interaction : :class:`~pincer.objects.app.interactions.Interaction`
        The interaction which is linked to the command.
    context : :class:`~pincer.objects.message.context.MessageContext`
        The context of the command.
    command : :class:`~pincer.utils.types.Coro`
        The coroutine which will be seen as a command.
    """
    sig, _ = get_signature_and_params(command)

    defaults = {
        key: value.default
        for key, value in sig.items()
        if value.default is not _empty
    }
    params = {}

    def get_options_from_command(options):
        if not options:
            return options
        if options[0].type == 1:
            return options[0].options
        if options[0].type == 2:
            return get_options_from_command(options[0].options)
        return options

    options = get_options_from_command(interaction.data.options)

    if options is not MISSING:
        params = {opt.name: opt.value for opt in options}

    args = []

    if interaction.data.type == AppCommandType.USER:
        # Add User and Member args
        user = next(iter(interaction.data.resolved.users.values()))

        if members := interaction.data.resolved.members:
            member = next(iter(members.values()))
            member.set_user_data(user)
            args.append(member)
        else:
            args.append(user)

    elif interaction.data.type == AppCommandType.MESSAGE:
        # Add Message to args
        args.append(next(iter(interaction.data.resolved.messages.values())))

    if interaction.data.values:
        args.append(interaction.data.values)

    kwargs = {**defaults, **params}

    try:
        await interaction_response_handler(
            self, command, manager, context, interaction, args, kwargs
        )
    except Exception as e:
        if coro := get_index(self.get_event_coro("on_command_error"), 0):
            try:
                await interaction_response_handler(
                    self,
                    coro.call,
                    coro.manager,
                    context,
                    interaction,
                    [e, *args],
                    kwargs,
                )
            except Exception as e:
                raise e
        raise e


async def interaction_create_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
) -> Tuple[str, Interaction]:
    """Middleware for ``on_interaction``, which handles command
    execution.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the interaction event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Raises
    ------
    e
        Generic try except on ``await interaction_handler`` and
        ``if 0 < len(params) < 3``

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.app.interactions.Interaction`]
        ``on_interaction_create`` and an ``Interaction``
    """
    interaction: Interaction = Interaction.from_dict(payload.data)
    call, manager = get_call(self, interaction)
    context = interaction.get_message_context()

    await interaction_handler(self, interaction, context, call, manager)

    return "on_interaction_create", interaction


def export() -> Coro:
    return interaction_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

import string
from typing import Union


class Color:
    """
    A color in RGB.

    Parameters
    ---------
    c : Union[:class:`str`,:class:`int`]
        The hex color can be a string that optionally starts with an ``#`` or
        an int with the RGB values.
    Attributes
    ---------
    r : :class:`int`
        The red value for this color.
    g : :class:`int`
        The green value for this color.
    b : :class:`int`
        The blue value for this color.
    """

    _MIN_VALUE = 0
    _MAX_VALUE = 1 << 24

    def __init__(self, c: Union[str, int]) -> None:
        if isinstance(c, int):
            if c < self._MIN_VALUE or c >= self._MAX_VALUE:
                raise ValueError("The color must be between 0 and 16777215")

            self.r = c >> 16 & 255
            self.g = c >> 8 & 255
            self.b = c & 255
            return

        # Conversion modified from this answer
        # https://stackoverflow.com/a/29643643
        if c.startswith("#"):
            c = c[1:]

        if len(c) != 6:
            raise ValueError("Hex value must be 6 characters.")

        if any(digit not in string.hexdigits for digit in c):
            raise ValueError("Illegal hex digit character.")

        self.r, self.g, self.b = (int(c[n:n + 2], 16) for n in (0, 2, 4))

    def __eq__(self, other):
        return self.r == other.r and self.g == other.g and self.b == other.b

    def __repr__(self):
        return f"Color(r={self.r}, g={self.g}, b={self.b})"

    def __str__(self) -> str:
        return f"#{self.hex}"

    @property
    def rbg(self):
        """
        Returns
        -------
        Tuple[:class:`int`, :class:`int` :class:`int`]
            Tuple value for rgb.
        """
        return self.r, self.g, self.b

    @property
    def hex(self):
        """
        Returns
        ---------
        str
            Str value for hex.
        """
        return f"{self.r:02x}{self.g:02x}{self.b:02x}"



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING, Optional

from ...utils.api_object import APIObject, ChannelProperty, GuildProperty

if TYPE_CHECKING:
    from ...client import Client
    from ...utils.snowflake import Snowflake


class PrivacyLevel(IntEnum):
    """Represents the level of publicity of a stage.

    Attributes
    ----------
    PUBLIC:
        The stage is public.
    GUILD_ONLY:
        The stage of for guild members only.
    """

    PUBLIC = 1
    GUILD_ONLY = 2


@dataclass(repr=False)
class StageInstance(APIObject, ChannelProperty, GuildProperty):
    """Represents a Stage Instance object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of this Stage instance
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Guild id of the associated Stage channel
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the associated Stage channel
    topic: :class:`str`
        Topic of the Stage instance (1-120 characters)
    privacy_level: :class:`~pincer.objects.guild.stage.PrivacyLevel`
        Privacy level of the Stage instance
    discoverable: :class:`bool`
        Is Stage Discovery enabled
    """

    id: Snowflake
    guild_id: Snowflake
    channel_id: Snowflake
    topic: str
    privacy_level: PrivacyLevel
    discoverable: bool

    @classmethod
    async def from_id(cls, client: Client, _id: int) -> StageInstance:
        return client.http.get(f"stage-instance/{_id}")

    async def modify(
        self,
        topic: Optional[str] = None,
        privacy_level: Optional[PrivacyLevel] = None,
        reason: Optional[str] = None,
    ):
        """|coro|
        Updates fields of an existing Stage instance.
        Requires the user to be a moderator of the Stage channel.

        Parameters
        ----------
        topic : Optional[:class:`str`]
            The topic of the Stage instance (1-120 characters)
        privacy_level : Optional[:class:`~pincer.objects.guild.stage.PrivacyLevel`]
            The privacy level of the Stage instance
        reason : Optional[:class:`str`]
            The reason for the modification
        """

        await self._client.modify_stage(self.id, topic, privacy_level, reason)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when the client gains access to a channel"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.thread import ThreadListSyncEvent

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def thread_list_sync(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_thread_list_sync`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the thread list sync event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.events.thread.ThreadListSyncEvent`]
        ``on_thread_list_sync`` and an ``ThreadListSyncEvent``
    """  # noqa: E501

    event = ThreadListSyncEvent.from_dict(payload.data)
    guild = self.guilds.get(event.guild_id)

    if guild:
        guild.threads = event.threads

    return "on_thread_list_sync", event


def export():
    return thread_list_sync



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .api_object import APIObject, ChannelProperty, GuildProperty
from .color import Color
from .conversion import remove_none
from .directory import chdir
from .event_mgr import EventMgr
from .extraction import get_index
from .insertion import should_pass_cls, should_pass_ctx
from .replace import replace
from .shards import calculate_shard_id
from .signature import get_params, get_signature_and_params
from .snowflake import Snowflake
from .tasks import Task, TaskScheduler
from .timestamp import Timestamp
from .types import (
    APINullable,
    Coro,
    MISSING,
    MissingType,
    choice_value_types,
    CheckFunction,
)

__all__ = (
    "APINullable", "APIObject", "ChannelProperty", "CheckFunction",
    "Color", "Coro", "EventMgr", "GuildProperty", "MISSING", "MissingType",
    "Snowflake", "Task", "TaskScheduler", "Timestamp", "calculate_shard_id",
    "chdir", "choice_value_types", "get_index", "get_params",
    "get_signature_and_params", "remove_none", "replace", "should_pass_cls",
    "should_pass_ctx"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .commands import command, user_command, message_command
from .chat_command_handler import ChatCommandHandler
from .arg_types import (
    CommandArg,
    Description,
    Choice,
    Choices,
    ChannelTypes,
    MaxValue,
    MinValue,
    Modifier,
)
from .components import (
    ActionRow, Button, ButtonStyle, ComponentHandler, SelectMenu, SelectOption,
    component, button, select_menu, LinkButton
)
from .groups import Group, Subgroup
from .interactable import Interactable, INTERACTION_REGISTERS

__all__ = (
    "ActionRow", "Button", "ButtonStyle", "ChannelTypes",
    "ChatCommandHandler", "Choice", "Choices", "CommandArg",
    "ComponentHandler", "Description", "Group", "INTERACTION_REGISTERS",
    "Interactable", "LinkButton", "MaxValue", "MinValue", "Modifier",
    "SelectMenu", "SelectOption", "Subgroup", "button", "command", "component",
    "message_command", "select_menu", "user_command"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from typing import Dict


from ._component import _Component
from ...utils.types import Singleton
from ...objects.app.command import InteractableStructure


class ComponentHandler(metaclass=Singleton):
    """Handles registered components

    Attributes
    ----------
    register : Dict[:class:`str`, :class:`Callable`]
        Dictionary of registered buttons.
    """

    register: Dict[str, InteractableStructure[_Component]] = {}



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, Generic, List, Optional, Union, TYPE_CHECKING, TypeVar

from .command_types import AppCommandOptionType, AppCommandType
from ..app.throttle_scope import ThrottleScope
from ...commands.groups import Group, Subgroup
from ...objects.guild.channel import ChannelType
from ...utils.api_object import APIObject, GuildProperty
from ...utils.snowflake import Snowflake
from ...utils.types import Coro, choice_value_types
from ...utils.types import MISSING

if TYPE_CHECKING:
    from ...utils.types import APINullable

T = TypeVar("T")

@dataclass(repr=False)
class AppCommandInteractionDataOption(APIObject):
    """Represents a Discord Application Command Interaction Data Option

    Attributes
    ----------
    name: :class:`str`
        The name of the parameter
    value: :class:`str`
        The value of the pair
    type: APINullable[:class:`str`]
        Value of application command option type
    options: APINullable[List[:data:`~pincer.objects.app.command.AppCommandInteractionDataOption`]]
        Present if this option is a group or subcommand
    """
    # noqa: E501
    name: str
    value: APINullable[str] = MISSING
    type: APINullable[AppCommandOptionType] = MISSING
    options: APINullable[
        List[AppCommandInteractionDataOption]] = MISSING


@dataclass(repr=False)
class AppCommandOptionChoice(APIObject):
    """Represents a Discord Application Command Option Choice object

    Attributes
    ----------
    name: :class:`str`
        1-100 character choice name
    value: Union[:data:`~pincer.utils.types.choice_value_types`]
        Value of the choice, up to 100 characters if string
    """
    name: str
    value: choice_value_types

    def __post_init__(self):
        # APIObject __post_init_ causes issues by converting `value` to a string
        self.name = str(self.name)


@dataclass(repr=False)
class AppCommandOption(APIObject):
    """Represents a Discord Application Command Option object

    Attributes
    ----------
    type: :class:`~pincer.objects.AppCommandOptionType`
        The type of option
    name: :class:`str`
        1-32 lowercase character name matching `^[\\w-]{1,32}$`
    description: :class:`str`
        1-100 character description
    required: APINullable[:class:`bool`]
        If the parameter is required or optional |default| :data:`False`
    choices: APINullable[List[:class:`~pincer.objects.app.command.AppCommandOptionChoice`]]
        Choices for ``STRING``, ``INTEGER``, and ``NUMBER``
        types for the user to pick from, max 25
    options: APINullable[List[:class:`~pincer.objects.app.command.AppCommandOptionChoice`]]
        If the option is a subcommand or subcommand group type,
        this nested options will be the parameters
    """
    # noqa: E501
    type: AppCommandOptionType
    name: str
    description: str

    required: bool = False
    autocomplete: APINullable[bool] = MISSING
    choices: APINullable[List[AppCommandOptionChoice]] = MISSING
    options: APINullable[List[AppCommandOption]] = MISSING
    channel_types: APINullable[List[ChannelType]] = MISSING
    min_value: APINullable[Union[int, float]] = MISSING
    max_value: APINullable[Union[int, float]] = MISSING


@dataclass(repr=False)
class AppCommand(APIObject, GuildProperty):
    """Represents a Discord Application Command object

    Attributes
    ----------
    type: :class:`~pincer.objects.app.command.AppCommandType`
        The type of command, defaults ``1`` if not set
    name: :class:`str`
        1-32 character name
    description: :class:`str`
        1-100 character description for ``CHAT_INPUT`` commands,
        empty string for ``USER`` and ``MESSAGE`` commands
    id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Unique id of the command
    version: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Auto-incrementing version identifier updated during substantial
        record changes
    application_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Unique id of the parent application
    options: APINullable[List[:class:`~pincer.objects.app.command.AppCommandOption`]]
        The parameters for the command, max 25
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Guild id of the command, if not global
    default_permission: APINullable[:class:`bool`]
        Whether the command is enabled by default
        when the app is added to a guild
    """
    # noqa: E501
    type: AppCommandType
    name: str
    description: str

    id: APINullable[Snowflake] = MISSING
    version: APINullable[Snowflake] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    options: APINullable[List[AppCommandOption]] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    default_permission: APINullable[bool] = True
    default_member_permissions: APINullable[None] = None
    dm_permission: APINullable[None] = None

    def __post_init__(self):
        super().__post_init__()

        if self.options is MISSING and self.type is AppCommandType.MESSAGE:
            self.options = []

    def __eq__(self, other: Union[AppCommand, InteractableStructure]):
        if isinstance(other, InteractableStructure):
            other = other.app

        # `description` and `options` are tested for equality with a custom check
        eq_props = (
            "type", "name", "guild_id", "default_permission", "options"
        )

        eq = (
            self.__getattribute__(prop) == other.__getattribute__(prop)
            for prop in eq_props
        )

        return all(
            (
                *eq,
                self.description == other.description
                # If this command has a MISSING description, Discord would return
                # registered the command with the description ''
                or self.description is MISSING and not other.description
            )
        )

    def __hash__(self):
        return hash((self.id, self.name, self.guild_id, self.type))

    def add_option(self, option: AppCommandOption):
        """Add a new option field to the current application command.

        Parameters
        ----------
        option : :class:`~pincer.objects.app.command.AppCommandOption`
            The option which will be appended.
        """
        if self.options:
            self.options.append(option)
        else:
            self.options = [option]


@dataclass(repr=False)
class InteractableStructure(Generic[T]):
    """Represents the structure of how the client saves the existing
    commands to registers. This is generic over Application Commands,
    Message Components, and Autocomplete.

    Attributes
    ----------
    call: :class:`~pincer.utils.types.Coro`
        The coroutine which should be called when the command gets
        executed.
    metadata: T
        The metadata for this command. |default| :data:`None`
    manager : Optional[Any]
        The manager for this interactable. |default| :data:`None`
    extensions: List[Callable[..., Awaitable[bool]]]
        List of extensions for this command. |default| :data:`[]`
    cooldown: :class:`int`
        Amount of times for cooldown |default| :data:`0`
    cooldown_scale: :class:`float`
        Search time for cooldown |default| :data:`60.0`
    cooldown_scope: :class:`~pincer.objects.app.throttle_scope.ThrottleScope`
        The type of cooldown |default| :data:`ThrottleScope.USER`
    """
    call: Coro

    metadata: Optional[T] = None
    manager: Optional[Any] = None
    extensions: List[Callable[..., Awaitable[bool]]] = field(default_factory=list)

    cooldown: int = 0
    cooldown_scale: float = 60.0
    cooldown_scope: ThrottleScope = ThrottleScope.USER

    group: APINullable[Group] = MISSING
    sub_group: APINullable[Subgroup] = MISSING

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.call(*args, **kwargs)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable
from typing import TYPE_CHECKING, Tuple, List
from typing import Union

from ..objects.app.interaction_flags import InteractionFlags
from ..objects.message.component import MessageComponent
from ..objects.message.embed import Embed
from ..objects.message.file import File
from ..objects.message.message import Message

PILLOW_IMPORT = True

try:
    from PIL.Image import Image
except (ModuleNotFoundError, ImportError):
    PILLOW_IMPORT = False

if TYPE_CHECKING:
    from pincer import Client

MessageConvertable = Union[Embed, Message, str, Tuple, List]


def convert_message(client: Client, message: MessageConvertable) -> Message:
    """
    Converts a message to a Message object.

    Parameters
    ----------
    client : :class:`~pincer.client.Client`
        The Client object for the bot
    message : MessageConvertable
        A value to be converted to a message

    Returns
    -------
    :class:`~pincer.objects.message.message.Message`
        The message object to be sent
    """
    if message and isinstance(message, Iterable) and not isinstance(message, str):
        kwargs = defaultdict(list)
        for item in message:
            list_to_message_dict(item, kwargs)
        message = Message(**kwargs)
    elif isinstance(message, Embed):
        message = Message(embeds=[message])
    elif PILLOW_IMPORT and isinstance(message, (File, Image)):
        message = Message(attachments=[message])
    elif not isinstance(message, Message):
        message = (
            Message(message)
            if message
            else Message(
                client.received_message,
                flags=InteractionFlags.EPHEMERAL
            )
        )
    return message


def list_to_message_dict(item, kwargs):
    if isinstance(item, Embed):
        kwargs["embeds"].append(item)
    elif isinstance(item, File) or (PILLOW_IMPORT and isinstance(item, Image)):
        kwargs["attachments"].append(item)
    elif isinstance(item, MessageComponent):
        kwargs["components"].append(item)
    elif isinstance(item, str):
        kwargs["content"] = item
    elif isinstance(item, InteractionFlags):
        kwargs["flags"] = item



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass
from typing import Optional

from ...objects.guild.role import Role
from ...objects.user.user import User


@dataclass(repr=False)
class Mentionable:
    """
    Represents the Mentionable type

    user : Optional[:class:`~pincer.objects.user.user.User`]
        User object returned from a discord interaction
    role: Optional[:class:`~pincer.objects.guild.role.Role`]
        Role object returned from a discord interaction
    """
    user: Optional[User] = None
    role: Optional[Role] = None

    @property
    def is_user(self):
        """Returns true if the Mentionable object has a User"""
        return self.user is not None

    @property
    def is_role(self):
        """Returns true if the Mentionable object has a Role"""
        return self.role is not None



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from typing import TYPE_CHECKING

import logging
from asyncio import gather

from ..utils.types import MISSING, Singleton

from ..exceptions import ForbiddenError
from ..objects.guild.guild import Guild
from ..objects.app.command import AppCommand, AppCommandOption
from ..objects.app.command_types import AppCommandOptionType, AppCommandType

if TYPE_CHECKING:
    from typing import List, Dict, Optional, ValuesView, Union
    from .interactable import Interactable
    from ..client import Client
    from ..utils.snowflake import Snowflake
    from ..objects.app.command import InteractableStructure

_log = logging.getLogger(__name__)


class ChatCommandHandler(metaclass=Singleton):
    """Singleton containing methods used to handle various commands

    The register and built_register
    -------------------------------
    I found the way Discord expects commands to be registered to be very different than
    how you want to think about command registration. i.e. Discord wants nesting but we
    don't want any nesting. Nesting makes it hard to think about commands and also will
    increase lookup time.
    The way this problem is avoided is by storing a version of the commands that we can
    deal with as library developers and a version of the command that Discord thinks we
    should provide. That is where the register and the built_register help simplify the
    design of the library.
    The register is simply where the "Pincer version" of commands gets saved to memory.
    The built_register is where the version of commands that Discord requires is saved.
    The register allows for O(1) lookups by storing commands in a Python dictionary. It
    does cost some memory to save two copies in the current iteration of the system but
    we should be able to drop the built_register in runtime if we want to. I don't feel
    that lost maintainability from this is optimal. We can index by in O(1) by checking
    the register but can still use the built_register if we need to do a nested lookup.

    Attributes
    ----------
    client: :class:`Client`
        The client object
    managers: Dict[:class:`str`, :class:`~typing.Any`]
        Dictionary of managers
    register: Dict[:class:`str`, :class:`~pincer.objects.app.command.InteractableStructure`[:class:`~pincer.objects.app.command.AppCommand`]]
        Dictionary of ``InteractableStructure``
    built_register: Dict[:class:`str`, :class:`~pincer.objects.app.command.AppCommand`]]
        Dictionary of ``InteractableStructure`` where the commands are converted to
        the format that Discord expects for sub commands and sub command groups.
    """  # noqa: E501

    has_been_initialized = False
    managers: List[Interactable] = []
    register: Dict[str, InteractableStructure[AppCommand]] = {}
    built_register: Dict[str, AppCommand] = {}

    # Endpoints:
    __get = "/commands"
    __delete = "/commands/{command.id}"
    __update = "/commands/{command.id}"
    __add = "/commands"
    __add_guild = "/guilds/{command.guild_id}/commands"
    __get_guild = "/guilds/{guild_id}/commands"
    __update_guild = "/guilds/{command.guild_id}/commands/{command.id}"
    __delete_guild = "/guilds/{command.guild_id}/commands/{command.id}"

    def __init__(self, client: Client):
        self.client = client
        self._api_commands: List[AppCommand] = []
        _log.debug(
            "%i commands registered.", len(ChatCommandHandler.register)
        )

        self.__prefix = f"applications/{self.client.bot.id}"

    async def get_commands(self) -> List[AppCommand]:
        """|coro|

        Get a list of app commands from Discord

        Returns
        -------
        List[:class:`~pincer.objects.app.command.AppCommand`]
            List of commands.
        """
        # TODO: Update if discord adds bulk get guild commands
        guild_commands = await gather(*(
            self.client.http.get(
                self.__prefix + self.__get_guild.format(
                    guild_id=guild.id if isinstance(guild, Guild) else guild
                )
            ) for guild in self.client.guilds
        ))
        return list(
            map(
                AppCommand.from_dict,
                await self.client.http.get(self.__prefix + self.__get)
                + [cmd for guild in guild_commands for cmd in guild],
            )
        )

    async def remove_command(self, cmd: AppCommand):
        """|coro|

        Remove a specific command

        Parameters
        ----------
        cmd : :class:`~pincer.objects.app.command.AppCommand`
            What command to delete
        """
        # TODO: Update if discord adds bulk delete commands
        if cmd.guild_id:
            _log.info(
                "Removing command `%s` with guild id %d from Discord",
                cmd.name,
                cmd.guild_id,
            )
        else:
            _log.info("Removing global command `%s` from Discord", cmd.name)

        remove_endpoint = self.__delete_guild if cmd.guild_id else self.__delete

        await self.client.http.delete(
            self.__prefix + remove_endpoint.format(command=cmd)
        )

    async def add_command(self, cmd: AppCommand):
        """|coro|

        Add an app command

        Parameters
        ----------
        cmd : :class:`~pincer.objects.app.command.AppCommand`
            Command to add
        """
        _log.info("Updated or registered command `%s` to Discord", cmd.name)

        add_endpoint = self.__add

        if cmd.guild_id:
            add_endpoint = self.__add_guild.format(command=cmd)

        await self.client.http.post(
            self.__prefix + add_endpoint, data=cmd.to_dict()
        )

    async def add_commands(self, commands: List[AppCommand]):
        """|coro|

        Add a list of app commands

        Parameters
        ----------
        commands : List[:class:`~pincer.objects.app.command.AppCommand`]
            List of command objects to add
        """
        await gather(*map(self.add_command, commands))

    @staticmethod
    def __build_local_commands():
        """Builds the commands into the format that Discord expects. See class info
        for the reasoning.
        """

        # Reset the built register
        ChatCommandHandler.built_register = {}

        for cmd in ChatCommandHandler.register.values():

            if cmd.sub_group:
                # If a command has a sub_group, it must be nested 2 levels deep.
                #
                # command
                #     subcommand-group
                #         subcommand
                #
                # The children of the subcommand-group object are being set to include
                # `cmd` If that subcommand-group object does not exist, it will be
                # created here. The same goes for the top-level command.
                #
                # First make sure the command exists. This command will hold the
                # subcommand-group for `cmd`.

                # `key` represents the hash value for the top-level command that will
                # hold the subcommand.
                key = _hash_app_command_params(
                    cmd.group.name,
                    cmd.metadata.guild_id,
                    AppCommandType.CHAT_INPUT,
                    None,
                    None,
                )

                if key not in ChatCommandHandler.built_register:
                    ChatCommandHandler.built_register[key] = AppCommand(
                        name=cmd.group.name,
                        description=cmd.group.description,
                        type=AppCommandType.CHAT_INPUT,
                        guild_id=cmd.metadata.guild_id,
                        options=[]
                    )

                # The top-level command now exists. A subcommand group now if placed
                # inside the top-level command. This subcommand group will hold `cmd`.

                children = ChatCommandHandler.built_register[key].options

                sub_command_group = AppCommandOption(
                    name=cmd.sub_group.name,
                    description=cmd.sub_group.description,
                    type=AppCommandOptionType.SUB_COMMAND_GROUP,
                    options=[]
                )

                # This for-else makes sure that sub_command_group will hold a reference
                # to the subcommand group that we want to modify to hold `cmd`

                for cmd_in_children in children:
                    if (
                        cmd_in_children.name == sub_command_group.name
                        and cmd_in_children.description == sub_command_group.description
                        and cmd_in_children.type == sub_command_group.type
                    ):
                        sub_command_group = cmd_in_children
                        break
                else:
                    children.append(sub_command_group)

                sub_command_group.options.append(AppCommandOption(
                    name=cmd.metadata.name,
                    description=cmd.metadata.description,
                    type=AppCommandOptionType.SUB_COMMAND,
                    options=cmd.metadata.options,
                ))

                continue

            if cmd.group:
                # Any command at this point will only have one level of nesting.
                #
                # Command
                #    subcommand
                #
                # A subcommand object is what is being generated here. If there is no
                # top level command, it will be created here.

                # `key` represents the hash value for the top-level command that will
                # hold the subcommand.

                key = _hash_app_command_params(
                    cmd.group.name,
                    cmd.metadata.guild_id,
                    AppCommandOptionType.SUB_COMMAND,
                    None,
                    None
                )

                if key not in ChatCommandHandler.built_register:
                    ChatCommandHandler.built_register[key] = AppCommand(
                        name=cmd.group.name,
                        description=cmd.group.description,
                        type=AppCommandOptionType.SUB_COMMAND,
                        guild_id=cmd.metadata.guild_id,
                        options=[]
                    )

                # No checking has to be done before appending `cmd` since it is the
                # lowest level.
                ChatCommandHandler.built_register[key].options.append(
                    AppCommandOption(
                        name=cmd.metadata.name,
                        description=cmd.metadata.description,
                        type=AppCommandType.CHAT_INPUT,
                        options=cmd.metadata.options
                    )
                )

                continue

            # All single-level commands are registered here.
            ChatCommandHandler.built_register[
                _hash_interactable_structure(cmd)
            ] = cmd.metadata

    @staticmethod
    def get_local_registered_commands() -> ValuesView[AppCommand]:
        return ChatCommandHandler.built_register.values()

    async def __get_existing_commands(self):
        """|coro|

        Get AppCommand objects for all commands registered to discord.
        """
        try:
            self._api_commands = await self.get_commands()
        except ForbiddenError:
            logging.error("Cannot retrieve slash commands, skipping...")
            return

    async def __remove_unused_commands(self):
        """|coro|

        Remove commands that are registered by discord but not in use
        by the current client
        """
        local_registered_commands = self.get_local_registered_commands()

        def should_be_removed(target: AppCommand) -> bool:
            # Commands have endpoints based on their `name` amd `guild_id`. Other
            # parameters can be updated instead of deleting and re-registering the
            # command.
            return all(
                target.name != reg_cmd.name
                and target.guild_id != reg_cmd.guild_id
                for reg_cmd in local_registered_commands
            )

        # NOTE: Cannot be generator since it can't be consumed due to lines 743-745
        to_remove = [*filter(should_be_removed, self._api_commands)]

        await gather(
            *(self.remove_command(cmd) for cmd in to_remove)
        )

        self._api_commands = [
            cmd for cmd in self._api_commands
            if cmd not in to_remove
        ]

    async def __add_commands(self):
        """|coro|
        Add all new commands which have been registered by the decorator to Discord.

        .. code-block::

            Because commands have unique names within a type and scope, we treat POST
            requests for new commands as upserts. That means making a new command with
            an already-used name for your application will update the existing command.
            `<https://discord.dev/interactions/application-commands#updating-and-deleting-a-command>`_

        Therefore, we don't need to use a separate loop for updating and adding
        commands.
        """
        for command in self.get_local_registered_commands():
            if command not in self._api_commands:
                await self.add_command(command)

    async def initialize(self):
        """|coro|

        Call methods of this class to refresh all app commands
        """
        if ChatCommandHandler.has_been_initialized:
            # Only first shard should be initialized.
            return

        ChatCommandHandler.has_been_initialized = True

        self.__build_local_commands()
        await self.__get_existing_commands()
        await self.__remove_unused_commands()
        await self.__add_commands()


def _hash_interactable_structure(interactable: InteractableStructure[AppCommand]):
    return _hash_app_command(
        interactable.metadata,
        interactable.group,
        interactable.sub_group
    )


def _hash_app_command(
    command: AppCommand,
    group: Optional[str],
    sub_group: Optional[str]
) -> int:
    """
    See :func:`~pincer.commands.commands._hash_app_command_params` for information.
    """
    return _hash_app_command_params(
        command.name,
        command.guild_id,
        command.type,
        group,
        sub_group
    )


def _hash_app_command_params(
    name: str,
    guild_id: Union[Snowflake, None, MISSING],
    app_command_type: AppCommandType,
    group: Optional[str],
    sub_group: Optional[str]
) -> int:
    """
    The group layout in Pincer is very different from what discord has on their docs.
    You can think of the Pincer group layout like this:

    name: The name of the function that is being called.

    group: The :class:`~pincer.commands.groups.Group` object that this function is
        using.
    sub_option: The :class:`~pincer.commands.groups.Subgroup` object that this
        functions is using.

    Abstracting away this part of the Discord API allows for a much cleaner
    transformation between what users want to input and what commands Discord
    expects.

    Parameters
    ----------
    name : str
        The name of the function for the command
    guild_id : Union[:class:`~pincer.utils.snowflake.Snowflake`, None, MISSING]
        The ID of a guild, None, or MISSING.
    app_command_type : :class:`~pincer.objects.app.command_types.AppCommandType`
        The app command type of the command. NOT THE OPTION TYPE.
    group : str
        The highest level of organization the command is it. This should always be the
        name of the base command. :data:`None` or :data:`MISSING` if not there.
    sub_option : str
        The name of the group that holds the lowest level of options. :data:`None` or
        :data:`MISSING` if not there.
    """
    return hash((name, guild_id, app_command_type, group, sub_group))



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from collections import ChainMap

from .. import client as _client
from .chat_command_handler import ChatCommandHandler
from .components.component_handler import ComponentHandler
from ..objects.app.command import AppCommand, InteractableStructure


INTERACTION_REGISTERS = ChainMap(ChatCommandHandler.register, ComponentHandler.register)


class Interactable:
    """
    Class that can register :class:`~pincer.commands.interactable.PartialInteractable`
    objects. Any class that subclasses this class can register Application Commands and
    Message Components.
    PartialInteractable objects are registered by running the register function and
    setting an attribute of the client to the result.
    """

    def __init__(self):
        for value in vars(type(self)).values():
            if isinstance(value, InteractableStructure):
                value.manager = self

    def __del__(self):
        self.unassign()

    def unassign(self):
        """
        Removes this objects loaded commands from ChatCommandHandler and
        ComponentHandler and removes loaded events from the client.
        """
        for value in vars(type(self)).values():
            if (
                isinstance(value, InteractableStructure)
                and isinstance(value.metadata, AppCommand)
            ):
                for key, _value in INTERACTION_REGISTERS.items():
                    if value is _value:
                        INTERACTION_REGISTERS.pop(key)

                key = value.call.__name__.lower()

                event_or_list = _client._events.get(key)
                if isinstance(event_or_list, list):
                    if value in event_or_list:
                        event_or_list.remove(value)
                else:
                    _client._events.pop(key, None)
